// xRxDBGrid Package  1998-1999, Deks
//
// Deks Site: http://deks.webjump.com/
// E-mail: Deks@online.nsk.su
// ============================================================================
// Main unit for package.

unit xRxDBGrid;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DB, DBTables, RxDBCtrl;

const
  // NOT for localize:
  xSQLPrefix = ' order by ';
  xSQLAsc = ' asc';
  xSQLDesc = ' desc';

  // maybe for localize (error messages only for developers):
  xSRG_Exception1 = 'TxRxDBGrid works with TRxQuery. Use TRxQuery instead TQuery, and define Macro for OrderBy SQL clause.';
  xSRG_Exception2 = 'Set SQLMacroName property for TRxQuery-based sortable grids.';

type
  TxOnGetSQLFieldName = procedure(Field: String; var SQLFieldName: String) of object;

  // ==================
  // TxRxDBGrid
  // ==================
  TxRxDBGrid = class(TRxDBGrid)
  private
  protected
    FSQLMacroName: String;
    FOnGetSQLFieldName: TxOnGetSQLFieldName;

    FSortFieldName: String;
    FSortDesc: Boolean;

    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; AOperation: TOperation); override;

    procedure GridGetBtnParams(Sender: TObject; Field: TField;
      AFont: TFont; var Background: TColor; var SortMarker: TSortMarker;
      IsDown: Boolean);
    procedure GridTitleBtnClick(Sender: TObject; ACol: Integer;
      Field: TField);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property SortFieldName: String read FSortFieldName;
    property SortDesc: Boolean read FSortDesc;
  published
    property SQLMacroName: String read FSQLMacroName write FSQLMacroName;
    property OnGetSQLFieldName: TxOnGetSQLFieldName read FOnGetSQLFieldName write FOnGetSQLFieldName;
  end;

implementation

uses RxQuery, DBUtils, StrUtils;

// ============================================================================
// TxRxDBGrid
// ============================================================================
constructor TxRxDBGrid.Create(AOwner: TComponent);
begin
  // Constructor for component.

  // call inherited constructor:
  inherited Create(AOwner);

  // init internal fields (variables):
  TitleButtons := True;
  FOnGetSQLFieldName := nil;
  FSortFieldName := '';
  FSortDesc := False;
end;

// ----------------------------------------------------------------------------
destructor TxRxDBGrid.Destroy;
begin
  // Destructor for component.

  // call inherited destructor:
  inherited Destroy;
end;

// ----------------------------------------------------------------------------
procedure TxRxDBGrid.Loaded;
begin
  // Handle "after loading" event for component.

  // call inherited method:
  inherited Loaded;

  // set event handler after loading component:
  if not (csDesigning in ComponentState) then begin
    OnGetBtnParams := GridGetBtnParams;
    OnTitleBtnClick := GridTitleBtnClick;
  end;
end;

// ----------------------------------------------------------------------------
procedure TxRxDBGrid.Notification(AComponent: TComponent; AOperation: TOperation);
begin
  // Handle changing to components in container (maybe, form).

  // call inherited method:
  inherited Notification(AComponent, AOperation);
end;

// ----------------------------------------------------------------------------
// handle display of sort marks on titles of grid:
procedure TxRxDBGrid.GridGetBtnParams(Sender: TObject; Field: TField;
  AFont: TFont; var Background: TColor; var SortMarker: TSortMarker;
  IsDown: Boolean);
begin
  if Field.FieldName = FSortFieldName then begin
    if FSortDesc then
      SortMarker := smUp  // if descending sort, then set Mark to "Up" state
    else
      SortMarker := smDown; // else marker is "Down"
  end;
end;

// ----------------------------------------------------------------------------
// handle clicks on title buttons:
procedure TxRxDBGrid.GridTitleBtnClick(Sender: TObject; ACol: Integer;
  Field: TField);
var
  aSortString: String;
  aSQLFieldName: String;
  aRxQuery: TRxQuery;
  aTable: TTable;
  i:Integer;
  isFound: Boolean;
begin
  // get dataset type:
  if DataSource=nil then Exit;
  if DataSource.DataSet=nil then Exit;

  if (DataSource.DataSet is TQuery) and not (DataSource.DataSet is TRxQuery)
  then raise Exception.Create(xSRG_Exception1);

  aRxQuery := nil;
  aTable := nil;
  if (DataSource.DataSet is TRxQuery) then
    aRxQuery := TRxQuery(DataSource.DataSet);
  if (DataSource.DataSet is TTable) then
    aTable := TTable(DataSource.DataSet);

  // change sort mark:
  if Field.FieldName = FSortFieldName then begin
    if aTable<>nil then begin
      // for TTable only ascending sort order is avail and second click will
      // toggle sorting to "none":
      FSortDesc := False;
      FSortFieldName := '';
    end else if (aRxQuery<>nil) then begin
      // change sort mark type for TQuery-based grids (3 positions):
      if FSortDesc = False then
        FSortDesc := True
      else if SortDesc = True then
        FSortFieldName := '';
    end;
  end else begin
    // set new sort field:

    // handle setting of new sort field for TTable-based DataSets:
    if aTable<> nil then begin
      // try to find new sort field in indexed field's list:
      isFound := False;
      FSortDesc := False;
      aTable.IndexDefs.Update;
      for i:=0 to aTable.IndexDefs.Count-1 do begin
        if aTable.IndexDefs[i].Fields = Field.FieldName then begin
          isFound := True;
          // Note: Paradox files can not handle descending indices as index
          // for sorting order (bug ?):
          if (ixDescending in aTable.IndexDefs[i].Options) or
             (ixExpression in aTable.IndexDefs[i].Options) then
             isFound := False;
        end;
      end;

      // is field is found, then we can sort by that field:
      if isFound = True then begin
        FSortFieldName := Field.FieldName;
      end;
   end else begin
      // for TQuery, we can sort by any field:
      FSortFieldName := Field.FieldName;
      FSortDesc := False;
    end;
  end;

  // apply new sort order for DataSet:
  if FSortFieldName <> '' then begin
    // ===== handle TQuery =====
    if aRxQuery<>nil then begin
      // create OredrBy clause for SQL statement:
      if Assigned(FOnGetSQLFieldName) then begin
        // use transformed FieldName:
        FOnGetSQLFieldName(SortFieldName, aSQLFieldName);
        aSortString := xSQLPrefix + aSQLFieldName;
      end else
        // use "raw" field name:
        aSortString := xSQLPrefix + SortFieldName;

      if SortDesc=False then
        aSortString := aSortString + xSQLAsc
      else
        aSortString := aSortString + xSQLDesc;

      // test. if SQLMacroName property is set:
      if FSQLMacroName = '' then
        raise Exception.Create(xSRG_Exception2);

      // send new OrderBy string to TQuery:
      aRxQuery.MacroByName(FSQLMacroName).Value := aSortString;

      // refresh query with new sort order:
      RefreshQuery(aRxQuery);
    end else
    // ==== handle TTable =====
    if (aTable <> nil) then begin
      // always set ascending sort order:
      //if FSortDesc=True then FSortDesc := False;

      // set sort order for table:
      aTable.IndexFieldNames := FSortFieldName;
    end;
  end else begin
    // remove sorting:
    if aRxQuery<>nil then begin
      aRxQuery.MacroByName(FSQLMacroName).Value := '';
      RefreshQuery(aRxQuery);
    end;
  end;
end;

end.
